/*===- TableGen'erated file -------------------------------------*- C++ -*-===*\
|*                                                                            *|
|* Property definitions for LLDB.                                             *|
|*                                                                            *|
|* Automatically generated file, do not edit!                                 *|
|* From: PlatformAndroidProperties.td                                         *|
|*                                                                            *|
\*===----------------------------------------------------------------------===*/

// Property definitions for android
#ifdef LLDB_PROPERTIES_android
static constexpr PropertyDefinition g_android_properties[] = {
  {"package-name", OptionValue::eTypeString, true, 0, "", {}, "Specify package name to run adb shell command with 'run-as' as the package user when necessary (e.g. to get file with 'cat' and 'dd')."},
};
#undef LLDB_PROPERTIES_android
#endif // android Property

