/*===- TableGen'erated file -------------------------------------*- C++ -*-===*\
|*                                                                            *|
|* Property definitions for LLDB.                                             *|
|*                                                                            *|
|* Automatically generated file, do not edit!                                 *|
|* From: SymbolFilePDBProperties.td                                           *|
|*                                                                            *|
\*===----------------------------------------------------------------------===*/

// Property definitions for symbolfilepdb
#ifdef LLDB_PROPERTIES_symbolfilepdb
static constexpr PropertyDefinition g_symbolfilepdb_properties[] = {
  {"reader", OptionValue::eTypeEnum, true, ePDBReaderDefault, nullptr, OptionEnumValues(g_pdb_reader_enums), "Selects the reader for PDB symbol files. The native PDB reader that uses LLVM's PDB support is always available (value: 'native'). Secondly, the DIA PDB reader is only available if LLVM was comppiled with Microsoft's DIA SDK on Windows (value: 'DIA'). By default, the DIA PDB reader is used if available. The LLDB_USE_NATIVE_PDB_READER environment variable can be used to switch to the native reader when this setting has the default value. Otherwise, the setting always has priority."},
};
#undef LLDB_PROPERTIES_symbolfilepdb
#endif // symbolfilepdb Property

